#%%
from astropy.io import fits
import importlib, os
import multiprocessing as mp
import numpy as np
import matplotlib.pyplot as plt
try:
    raise Exception("Force except block")
    import ACID_code_v2 as acid
except:
    os.chdir(os.path.dirname(__file__))
    os.chdir("../src") # ensures we are in the main directory
    import ACID_code_v2 as acid
    os.chdir("../example") # change back to example directory
importlib.reload(acid)

spec_file = fits.open('sample_spec_1.fits')

wavelength = spec_file[0].data   # Wavelengths in Angstroms
spectrum = spec_file[1].data     # Spectral Flux
error = spec_file[2].data        # Spectral Flux Errors
sn = spec_file[3].data           # SN of Spectrum

linelist = 'example_linelist.txt' # Insert path to line list

# choose a velocity grid for the final profile(s)
deltav = 0.82   # velocity pixel size must not be smaller than the spectral pixel size
velocities = np.arange(-25, 25, deltav)

# run ACID function
result = acid.ACID([wavelength], [spectrum], [error], linelist, [sn], velocities,
                   parallel=True, cores=None, nsteps=4000, verbose=True)

# extract profile and errors
profile = result[0, 0, 0]
profile_error = result[0, 0, 1]

# plot results
plt.figure()
plt.errorbar(velocities, profile, profile_error)
plt.xlabel('Velocities (km/s)')
plt.ylabel('Flux')
plt.show()


import numpy as np

def model_func(inputs, x, k_max, alpha):
    ## model for the mcmc - takes the profile(z) and the continuum coefficents(inputs[k_max:]) to create a model spectrum.
    z = inputs[:k_max]

    mdl = np.dot(alpha, z) ##alpha has been declared a global variable after LSD is run.

    #converting model from optical depth to flux
    mdl = np.exp(mdl)

    ## these are used to adjust the wavelengths to between -1 and 1 - makes the continuum coefficents smaller and easier for emcee to handle.
    a = 2/(np.max(x)-np.min(x))
    b = 1 - a*np.max(x)

    mdl1 = 0
    for i in range(k_max, len(inputs) - 1):
        mdl1 = mdl1 + (inputs[i] * ((x * a) + b) ** (i - k_max))

    # coefs = np.asarray(inputs[k_max:-1], dtype=float) # Potential improvement - Ben
    # X = (a * x) + b
    # if coefs.size:
    #     powers = np.arange(coefs.size)
    #     # X[:, None] ** powers -> shape (len(x), coefs.size); dot with coefs -> (len(x),)
    #     mdl1 = np.dot(X[:, None] ** powers[None, :], coefs)
    # else:
    #     mdl1 = 0.0

    mdl1 = mdl1 * inputs[-1]
    
    mdl = mdl * mdl1
   
    return mdl

class Model:
    def __init__(self, x, y, yerr, velocities, k_max, alpha):
        self.x = x
        self.y = y
        self.yerr = yerr
        self.velocities = velocities
        self.k_max = k_max
        self.alpha = alpha

    def log_likelihood(self, theta, x, y, yerr):
        ## maximum likelihood estimation for the mcmc model.
        model = model_func(theta, x, k_max=self.k_max, alpha=self.alpha)

        lnlike = -0.5 * np.sum(((y) - (model)) ** 2 / yerr**2 + np.log(yerr**2)+ np.log(2*np.pi))

        return lnlike

    def log_prior(self, theta):
        ## imposes the prior restrictions on the inputs - rejects if profile point is less than -10 or greater than 0.5.
        check = 0
        z = theta[:self.k_max]

        for i in range(len(theta)):
            if i < self.k_max: ## must lie in z
                if -10 <= theta[i] <= 0.5:
                    pass
                else:
                    check = 1

        if check == 0:

            # excluding the continuum points in the profile (in flux)
            z_cont = []
            v_cont = []
            for i in range(0, 5):
                    z_cont.append(np.exp(z[len(z)-i-1])-1)
                    v_cont.append(self.velocities[len(self.velocities)-i-1])
                    z_cont.append(np.exp(z[i])-1)
                    v_cont.append(self.velocities[i])

            z_cont = np.array(z_cont)
            v_cont = np.array(v_cont)

            p_pent = np.sum((np.log((1/np.sqrt(2*np.pi*0.01**2)))-0.5*(z_cont/0.01)**2))

            return p_pent

        return -np.inf

    def log_probability(self, theta):
        ## calculates log probability - used for mcmc
        lp = self.log_prior(theta)
        if not np.isfinite(lp):
            return -np.inf
        final = lp + self.log_likelihood(theta, self.x, self.y, self.yerr)
        return final


import numpy as np
from math import log10, floor
import glob

def ensure_list(x, allow_none=False):
    # Ensure inputs are lists, np.arrays are converted to lists
    if isinstance(x, list):
        return x
    if x is None:
        if allow_none:
            return None
        else:
            raise TypeError("Input must be a list or numpy array, not None")
    if isinstance(x, (str, bytes, bytearray)):
        return [x]
    if isinstance(x, np.ndarray):
        if x.ndim == 0:
            return [x.tolist()]
        else:
            return x.tolist()
    else:
        raise TypeError("Input must be a list or numpy array")

def round_sig(x1, sig):
    return round(x1, sig-int(floor(log10(abs(x1))))-1)

def findfiles(directory, file_type):

    filelist_corrected = glob.glob('%s/*/*%s**A_corrected*.fits'%(directory, file_type)) #finding corrected spectra
    filelist = glob.glob('%s/*/*%s**A*.fits'%(directory, file_type)) #finding all A band spectra

    filelist_final = []
    filelist_corrected_set = set(filelist_corrected)

    for file in filelist: #filtering out corrected spectra
        if file not in filelist_corrected_set:
            filelist_final.append(file)

    return filelist_final

def od2flux(x):
    return np.exp(x)-1

def flux2od(x):
    return np.log(x+1)


#%%
from astropy.io import fits
import os, glob, importlib
os.chdir(os.path.dirname(__file__))
os.chdir("..") # ensures we are in the main directory
import ACID_code_v2 as acid
import numpy as np
import matplotlib.pyplot as plt
importlib.reload(acid)

def test_run_e2ds():

    e2ds_files = glob.glob('tests/data/*e2ds_A*.fits')
    linelist = 'example/example_linelist.txt'
    save_path = 'no save'

    velocities = np.arange(-25, 25, 0.82)

    # run ACID on e2ds files
    ACID_results_e2ds = acid.ACID_HARPS(e2ds_files, linelist, vgrid = velocities, save_path = save_path,
                                        order_range = np.arange(41, 43), nsteps=2000)

def test_run_s1d():

    s1d_files = glob.glob('tests/data/*s1d_A*.fits')
    linelist = 'example/example_linelist.txt'
    save_path = 'no save'

    velocities = np.arange(-25, 25, 0.82)

    # run ACID on s1d files
    ACID_results_s1d = acid.ACID_HARPS(s1d_files, linelist, vgrid = velocities, save_path = save_path,
                                       order_range = np.arange(41, 43), file_type = 's1d', nsteps=2000)

test_run_e2ds()
test_run_s1d()
print("All tests passed!")



import sys, emcee, warnings, os, time, importlib
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from scipy.interpolate import interp1d
import multiprocessing as mp
from functools import partial
from multiprocessing import Pool
from . import utils
from . import LSD
from . import mcmc_utils

warnings.filterwarnings("ignore")
importlib.reload(LSD)
importlib.reload(utils)


def continuumfit(fluxes, wavelengths, errors, poly_ord):
        
        cont_factor = fluxes[0]
        if cont_factor == 0: 
            cont_factor = np.mean(fluxes)
        idx = wavelengths.argsort()
        wavelength = wavelengths[idx]
        fluxe = fluxes[idx] / cont_factor
        clipped_flux = []
        clipped_waves = []
        binsize = 100
        for i in range(0, len(wavelength), binsize):
            waves = wavelength[i:i+binsize]
            flux = fluxe[i:i+binsize]
            indicies = flux.argsort()
            flux = flux[indicies]
            waves = waves[indicies]
            clipped_flux.append(flux[len(flux)-1])
            clipped_waves.append(waves[len(waves)-1])
        coeffs = np.polyfit(clipped_waves, clipped_flux, poly_ord)
        poly = np.poly1d(coeffs)
        fit = poly(wavelengths) * cont_factor
        flux_obs = fluxes / fit
        new_errors = errors / fit

        return np.concatenate((np.flip(coeffs), [cont_factor])), flux_obs, new_errors, fit

def read_in_frames(order, filelist, file_type, directory=None):
    
    # read in first frame
    fluxes, wavelengths, flux_error_order, sn, mid_wave_order, telluric_spec, overlap = LSD.blaze_correct(
        file_type, 'order', order, filelist[0], directory, 'unmasked', run_name, 'y')
    
    frames = np.zeros((len(filelist), len(wavelengths)))
    errors = np.zeros((len(filelist), len(wavelengths)))
    frame_wavelengths = np.zeros((len(filelist), len(wavelengths)))
    sns = np.zeros((len(filelist), ))

    frames[0] = fluxes
    errors[0] = flux_error_order
    frame_wavelengths[0] = wavelengths
    sns[0] = sn

    def task_frames(frames, errors, frame_wavelengths, sns, i):
        file = filelist[i]
        frames[i], frame_wavelengths[i], errors[i], sns[i], mid_wave_order, telluric_spec, overlap = LSD.blaze_correct(file_type, 'order', order, file, directory, 'unmasked', run_name, 'y')
        # print(i, frames)
        return frames, frame_wavelengths, errors, sns
    
    ### reads in each frame and corrects for the blaze function, adds the spec, errors and sn to their subsequent lists
    for i in range(len(filelist[1:])+1):
        # print(i)
        frames, frame_wavelengths, errors, sns = task_frames(frames, errors, frame_wavelengths, sns, i)
        
    ### finding highest S/N frame, saves this as reference frame

    idx = (sns==np.max(sns))
    global reference_wave
    reference_wave = frame_wavelengths[idx][0]
    reference_frame = frames[idx][0]
    reference_frame[reference_frame == 0] = 0.001
    reference_error = errors[idx][0]
    reference_error[reference_frame == 0] = 1000000000000000000

    global frames_unadjusted
    frames_unadjusted = frames
    global frame_errors_unadjusted
    frame_errors_unadjusted = errors

    ### each frame is divided by reference frame and then adjusted so that all spectra lie at the same continuum
    for n in range(len(frames)):
        f2 = interp1d(frame_wavelengths[n], frames[n], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
        div_frame = f2(reference_wave)/reference_frame

        idx_ref = (reference_frame<=0)
        div_frame[idx_ref]=1

        binned = []
        binned_waves = []
        binsize = int(round(len(div_frame)/5, 1))
        for i in range(0, len(div_frame), binsize):
            if i+binsize<len(reference_wave):
                waves = reference_wave[i:i+binsize]
                flux = div_frame[i:i+binsize]
                waves = waves[abs(flux-np.median(flux))<0.1]
                flux = flux[abs(flux-np.median(flux))<0.1]
                binned.append(np.median(flux))
                binned_waves.append(np.median(waves))

        binned = np.array(binned)
        binned_waves = np.array(binned_waves)
    
        ### fitting polynomial to div_frame
        try:coeffs = np.polyfit(binned_waves, binned, 4)
        except:coeffs = np.polyfit(binned_waves, binned, 2)
        poly = np.poly1d(coeffs)
        fit = poly(frame_wavelengths[n])
        frames[n] = frames[n]/fit
        errors[n] = errors[n]/fit
        idx = (frames[n] == 0)
        frames[n][idx] = 0.00001
        errors[n][idx] = 1000000000

    return frame_wavelengths, frames, errors, sns, telluric_spec

def calc_deltav(wavelengths):
    """Calculates velocity pixel size

    Calculates the velocity pixel size for the LSD velocity grid based off the spectral wavelengths.

    Args:
        wavelengths (array): Wavelengths for ACID input spectrum (in Angstroms).
        
    Returns:
        float: Velocity pixel size in km/s
    """ 
    resol1 = (wavelengths[-1]-wavelengths[0])/len(wavelengths)
    return resol1 / (wavelengths[0]+((wavelengths[-1]-wavelengths[0])/2)) * LSD.ckms

def combine_spec(wavelengths_f, spectra_f, errors_f, sns_f):

    interp_spec = np.zeros(spectra_f.shape)
    # combine all spectra to one spectrum
    for n in range(len(wavelengths_f)):

        global reference_wave
        idx_ref = (sns_f == np.max(sns_f))
        reference_wave = wavelengths_f[idx_ref][0]

        idx = np.where(wavelengths_f[n] != 0)[0]

        f2 = interp1d(wavelengths_f[n][idx], spectra_f[n][idx], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
        f2_err = interp1d(wavelengths_f[n][idx], errors_f[n][idx], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
        spectra_f[n] = f2(reference_wave)
        errors_f[n] = f2_err(reference_wave)

        # print(spectra_f[n])
        # print(errors_f[n])

        ## mask out out extrapolated areas
        idx_ex = np.logical_and(reference_wave<=np.max(wavelengths_f[n][idx]), reference_wave>=np.min(wavelengths_f[n][idx]))
        idx_ex = tuple([idx_ex==False])

        spectra_f[n][idx_ex] = 1.
        errors_f[n][idx_ex] = 1000000000000

        ## mask out nans and zeros (these do not contribute to the main spectrum)
        where_are_NaNs = np.isnan(spectra_f[n])
        errors_f[n][where_are_NaNs] = 1000000000000
        where_are_zeros = np.where(spectra_f[n] == 0)[0]
        errors_f[n][where_are_zeros] = 1000000000000

        where_are_NaNs = np.isnan(errors_f[n])
        errors_f[n][where_are_NaNs] = 1000000000000
        where_are_zeros = np.where(errors_f[n] == 0)[0]
        errors_f[n][where_are_zeros] = 1000000000000

    width = len(reference_wave)
    spectrum_f = np.zeros((width,))
    spec_errors_f = np.zeros((width,))

    for n in range(0,width):
        temp_spec_f = spectra_f[:, n]
        temp_err_f = errors_f[:, n]

        weights_f = (1/temp_err_f**2)

        idx = tuple([temp_err_f>=1000000000000])
        # print(weights_f[idx])
        weights_f[idx] = 0.

        if sum(weights_f) > 0:
            weights_f = weights_f / np.sum(weights_f)

            spectrum_f[n] = sum(weights_f * temp_spec_f)
            sn_f = sum(weights_f * sns_f) / sum(weights_f)

            spec_errors_f[n] = 1 / (sum(weights_f ** 2))
        
        else: 
            spectrum_f[n] = np.mean(temp_spec_f)
            spec_errors_f[n] = 1000000000000

    
    return reference_wave, spectrum_f, spec_errors_f, sn_f

def residual_mask(wavelengths, data_spec_in, data_err, initial_inputs, poly_ord, linelist,
                  velocities=np.arange(-25, 25, 0.82), pix_chunk=20, dev_perc=25, 
                  tell_lines=[3820.33, 3933.66, 3968.47, 4327.74, 4307.90, 4383.55,
                              4861.34, 5183.62, 5270.39, 5889.95, 5895.92, 6562.81,
                              7593.70, 8226.96], n_sig=1):
    ## iterative residual masking - mask continuous areas first - then possibly progress to masking the narrow lines

    forward = mcmc_utils.model_func(initial_inputs, wavelengths, k_max=k_max, alpha=alpha)

    a = 2 / (np.max(wavelengths) - np.min(wavelengths))
    b = 1 - a * np.max(wavelengths)

    mdl1 = 0
    for i in range(k_max, len(initial_inputs) - 1):
        mdl1 = mdl1 + (initial_inputs[i] * ((wavelengths * a) + b) ** (i - k_max))

    mdl1 = mdl1 * initial_inputs[-1]

    data_normalised = (data_spec_in - np.min(data_spec_in)) / (np.max(data_spec_in) - np.min(data_spec_in))
    forward_normalised = (forward - np.min(forward)) / (np.max(forward) - np.min(forward))
    residuals = data_normalised - forward_normalised

    data_err_compare = data_err.copy()
    
    ### finds consectuative sections where at least pix_chunk points have residuals greater than 0.25 - these are masked
    idx = (abs(residuals) > dev_perc / 100)

    flag_min = 0
    flag_max = 0
    for value in range(len(idx)):
        if idx[value] == True and flag_min <= value:
            flag_min = value
            flag_max = value
        elif idx[value] == True and flag_max < value:
            flag_max = value
        elif idx[value] == False and flag_max - flag_min >= pix_chunk:
            data_err[flag_min:flag_max] = 10000000000000000000
            flag_min = value
            flag_max = value

    ##############################################
    #                  TELLURICS                 #   
    ##############################################

    # data_err_compare = data_err.copy()

    ## masking tellurics
    for line in tell_lines:
        limit = (21/2.99792458e5)*line +3
        idx = np.logical_and((line-limit) <= wavelengths, wavelengths <= (limit+line))
        data_err[idx] = 10000000000000000000

    residual_masks = tuple([data_err >= 10000000000000000000])

    ###################################
    ###      sigma clip masking     ###
    ###################################

    m = np.median(residuals)
    sigma = np.std(residuals)
    a = 1

    upper_clip = m + a * sigma
    lower_clip = m - a * sigma

    rcopy = residuals.copy()

    idx1 = tuple([rcopy <= lower_clip])
    idx2 = tuple([rcopy >= upper_clip])

    data_err[idx1] = 10000000000000000000
    data_err[idx2] = 10000000000000000000

    poly_inputs, bin, bye, fit = continuumfit(data_spec_in, (wavelengths * a) + b, data_err, poly_ord)
    velocities1, profile, profile_err, alpha, continuum_waves, continuum_flux, no_line = LSD.LSD(
        wavelengths, bin, bye, linelist, 'False', poly_ord, 100, 30, run_name, velocities)

    # ## comment if you would like to keep sigma clipping masking in for final LSD run 
    # residual_masks = tuple([data_err>=1000000000000000000])

    return data_err, np.concatenate((profile, poly_inputs)), residual_masks

def get_profiles(all_frames, order, poly_cos, continuum_error, counter):
    flux = frames[counter]
    error = frame_errors[counter]
    wavelengths = frame_wavelengths[counter]
    sn = sns[counter]
   
    a = 2/(np.max(wavelengths)-np.min(wavelengths))
    b = 1 - a*np.max(wavelengths)

    mdl1 =0
    for i in np.arange(0, len(poly_cos)-1):
        mdl1 = mdl1+poly_cos[i]*((a*wavelengths)+b)**(i)
    mdl1 = mdl1*poly_cos[-1]

    #masking based off residuals interpolated onto new wavelength grid
    if len(frame_wavelengths)>1:
        reference_wave = frame_wavelengths[sns==max(sns)][0]
    else:
        reference_wave = frame_wavelengths[0]
    mask_pos = np.ones(reference_wave.shape)
    mask_pos[mask_idx]=10000000000000000000
    f2 = interp1d(reference_wave, mask_pos, bounds_error = False, fill_value = np.nan)
    interp_mask_pos = f2(wavelengths)
    interp_mask_idx = tuple([interp_mask_pos>=10000000000000000000])

    error[interp_mask_idx]=10000000000000000000

    # corrrecting continuum
    error = (error/flux) + (continuum_error/mdl1)
    flux = flux/mdl1
    error  = flux*error

    remove = tuple([flux<0])
    flux[remove]=1.
    error[remove]=10000000000000000000

    idx = tuple([flux>0])
    
    if len(flux[idx])==0:
        print('continuing... frame %s'%counter)
    
    else:
        velocities1, profile1, profile_errors, alpha, continuum_waves, continuum_flux, no_line= LSD.LSD(
            wavelengths, flux, error, linelist, 'False', poly_ord, sn, 10, 'test', velocities)

        p = np.exp(profile1)-1

        profile_f = np.exp(profile1)
        profile_errors_f = np.sqrt(profile_errors**2/profile_f**2)
        profile_f = profile_f-1

        all_frames[counter, order]=[profile_f, profile_errors_f]
        
        return all_frames

def combineprofiles(spectra, errors):
    spectra = np.array(spectra)
    idx = np.isnan(spectra)
    shape_og = spectra.shape
    if len(spectra[idx])>0:
        spectra = spectra.reshape((len(spectra)*len(spectra[0]), ))
        for n in range(len(spectra)):
            if spectra[n] == np.nan:
                spectra[n] = (spectra[n+1]+spectra[n-1])/2
                if spectra[n] == np.nan:
                    spectra[n] = 0.
    spectra = spectra.reshape(shape_og)
    errors = np.array(errors)

    
    spectra_to_combine = []
    weights=[]
    for n in range(0, len(spectra)):
        if np.sum(spectra[n])!=0:
            spectra_to_combine.append(list(spectra[n]))
            temp_err = np.array(errors[n, :])
            weight = (1/temp_err**2)
            weights.append(np.mean(weight))
    weights = np.array(weights/sum(weights))

    spectra_to_combine = np.array(spectra_to_combine)

    length, width = np.shape(spectra_to_combine)
    spectrum = np.zeros((1,width))
    spec_errors = np.zeros((1,width))

    for n in range(0, width):
        temp_spec = spectra_to_combine[:, n]
        spectrum[0,n]=sum(weights*temp_spec)/sum(weights)
        spec_errors[0,n] = (np.std(temp_spec, ddof=1)**2) * np.sqrt(sum(weights**2))

    spectrum = list(np.reshape(spectrum, (width,)))
    spec_errors = list(np.reshape(spec_errors, (width,)))

    return  spectrum, spec_errors

def ACID(input_wavelengths, input_spectra, input_spectral_errors, line, frame_sns,
         vgrid, all_frames='default', poly_or=3, pix_chunk=20, dev_perc=25, name="test",
         n_sig=1, telluric_lines=None, order=0, verbose=True, parallel=True, cores=None,
         nsteps=8000):
    """Accurate Continuum fItting and Deconvolution

    Fits the continuum of the given spectra and performs LSD on the continuum corrected spectra, returning an LSD profile for each spectrum given. 
    Spectra must cover a similiar wavelength range.

    Parameters
    ----------
    input_wavelengths : list or array
        Wavelengths for each frame (in Angstroms).
    input_spectra : list or array
        Spectral frames (in flux).
    input_spectral_errors : list or array
        Errors for each frame (in flux).
    line : str
        Path to linelist. Takes VALD linelist in long or short format as input. Minimum line depth input into VALD must
        be less than 1/(3*SN) where SN is the highest signal-to-noise ratio of the spectra. 
    frame_sns : list or array
        Average signal-to-noise ratio for each frame (used to calculate minimum line depth to consider from line list. 
    vgrid : array
        Velocity grid for LSD profiles (in km/s).
    all_frames : str or array, optional
        Output array for resulting profiles. Only neccessary if looping ACID function over many wavelength
        regions or order (in the case of echelle spectra). General shape needs to be (no. of frames, 1, 2, no. of velocity pixels)., by default 'default'
    poly_or : int, optional
        Order of polynomial to fit as the continuum, by default 3
    pix_chunk : int, optional
        Size of 'bad' regions in pixels. 'bad' areas are identified by the residuals between an inital model
        and the data. If a residual deviates by a specified percentage (dev_perc) for a specified number of pixels, by default 20
    dev_perc : int, optional
        Allowed deviation percentage. 'bad' areas are identified by the residuals between an inital model
        and the data. If a residual deviates by a specified percentage (dev_perc) for a specified number of pixels, by default 25
    n_sig : int, optional
        Number of sigma to clip in sigma clipping. Ill fitting lines are identified by sigma-clipping the
        residuals between an inital model and the data. The regions that are clipped from the residuals will
        be masked in the spectra. This masking is only applied to find the continuum fit and is removed when
        LSD is applied to obtain the final profiles, by default 1
    telluric_lines : list, optional
        List of wavelengths (in Angstroms) of telluric lines to be masked. This can also include problematic
        lines/features that should be masked also. For each wavelengths in the list ~3Å eith side of the line is masked., by default None
    order : int, optional
        Only applicable if an all_frames output array has been provided as this is the order position in that
        array where the result should be input. i.e. if order = 5 the output profile and errors would be inserted in all_frames[:, 5]., by default 0
    verbose : bool, optional
        If True prints out time taken for each section of the code. Defaults to True., by default True
    parallel : bool, optional
        If True uses multiprocessing to calculate the profiles for each frame in parallel, by default True
    cores : int, optional
        Number of cores to use if parallel=True. If None (default) all available cores will be used, by default None
    nsteps : int, optional
        nsteps (int, optional): Number of steps for the MCMC to run, try increasing if it doesn't converge, by default 8000

    Returns
    -------
    array
        Resulting profiles and errors for spectra.

    Raises
    ------
    TypeError
        If the input types are not as expected.
    """
    ### Setup

    # Ensure inputs are lists, np.arrays are converted to lists
    input_wavelengths, input_spectra, input_spectral_errors, frame_sns = [
        utils.ensure_list(v) for v in (input_wavelengths, input_spectra, input_spectral_errors, frame_sns)]

    # Define tell_lines if not input, check type if it is
    if telluric_lines is None:
        telluric_lines = [3820.33, 3933.66, 3968.47, 4327.74, 4307.90, 4383.55, 4861.34,
                          5183.62, 5270.39, 5889.95, 5895.92, 6562.81, 7593.70, 8226.96]
    elif not (isinstance(telluric_lines, list) or isinstance(telluric_lines, int)):
        raise TypeError("telluric_lines must be a list or integer")

    global velocities
    velocities = vgrid.copy()
    global linelist
    linelist = line
    global poly_ord
    poly_ord = poly_or
    global run_name
    run_name = name
    global frames
    global frame_wavelengths
    global frame_errors
    global sns
    frame_wavelengths = np.array(input_wavelengths)
    frames = np.array(input_spectra)
    frame_errors = np.array(input_spectral_errors)
    sns = np.array(frame_sns)

    if type(all_frames) != np.ndarray:
        if all_frames == 'default':
            all_frames = np.zeros((len(frames), 1, 2, len(velocities)))

    fw = frame_wavelengths.copy()
    f = frames.copy()
    fe = frame_errors.copy()
    s = sns.copy()

    if verbose:
        t0 = time.time()
        print('Initialising...')

    ## combines spectra from each frame (weighted based of S/N), returns to S/N of combined spec
    if len(fw)>1:
        wavelengths, fluxes, flux_error_order, sn = combine_spec(fw, f, fe, s)
    else:
        wavelengths, fluxes, flux_error_order, sn = fw[0], f[0], fe[0], s[0]

    ### getting the initial polynomial coefficents
    a = 2 / (np.max(wavelengths)-np.min(wavelengths))
    b = 1 - a * np.max(wavelengths)
    poly_inputs, fluxes1, flux_error_order1, fit = continuumfit(
        fluxes, (wavelengths*a)+b, flux_error_order, poly_ord)

    # if verbose:
    #     t2 = time.time()
    #     print('Set up before LSD %s'%(t2-t0))

    #### getting the initial profile
    global alpha
    velocities, profile, profile_errors, alpha, continuum_waves, continuum_flux, no_line = LSD.LSD(
        wavelengths, fluxes1, flux_error_order1, linelist, 'False', poly_ord, sn, 30, run_name,
        velocities, verbose=verbose)

    # if verbose:
    #     t3 = time.time()
    #     print('LSD run takes: %s'%(t3-t2))

    ## Setting the number of points in vgrid (k_max)
    global k_max
    k_max = len(profile)
    model_inputs = np.concatenate((profile, poly_inputs))

    ## Setting x, y, yerr for emcee
    global x, y, yerr
    x = wavelengths
    y = fluxes
    yerr = flux_error_order

    ## Setting these normalisation factors as global variables - used in the figures below
    a = 2 / (np.max(x) - np.min(x))
    b = 1 - a * np.max(x)

    # Masking based off residuals
    global mask_idx
    
    if verbose:
        print('Residual masking...')
    yerr, model_inputs_resi, mask_idx = residual_mask(x, y, yerr, model_inputs, poly_ord,
                                                      linelist, pix_chunk=pix_chunk,
                                                      dev_perc=dev_perc, tell_lines=telluric_lines,
                                                      n_sig=n_sig)

    # if verbose:
    #     t4 = time.time()
    #     print('residual masking takes: %s' %(t4-t3))

    ## Setting number of walkers and their start values(pos)
    ndim = len(model_inputs)
    nwalkers = ndim * 3
    rng = np.random.default_rng()

    ### starting values of walkers with independent variation
    sigma = 0.8 * 0.005
    pos = []
    for i in range(0, ndim):
        if i < ndim - poly_ord - 2:
            pos2 = rng.normal(model_inputs[i], sigma, (nwalkers, ))
        else:
            sigma = abs(utils.round_sig(model_inputs[i], 1)) / 10
            pos2 = rng.normal(model_inputs[i], sigma, (nwalkers, ))
        pos.append(pos2)

    pos = np.array(pos)
    pos = np.transpose(pos)

    # Initialise MCMC class:
    mcmc_class = mcmc_utils.Model(x, y, yerr, velocities, k_max, alpha)

    if verbose:
        t5 = time.time()
        # print('MCMC set up takes: %s'%(t5-t4))
        print('Initialised in %ss'%round((t5-t0), 2))

    print('Fitting the continuum using emcee...')

    if parallel:

        # Choose a sensible number of workers
        if cores is None:
            cores = os.cpu_count()  # Use all available cores

        if verbose:
            print(f"Using {cores} out of {os.cpu_count()} cores for MCMC")

        # For some reason, unspecified pooling as was before (as in case of windows in the else statement)
        # leds to a hung computer. So specify mp.get_context required, default is spawn, but spawn
        # causes multiple instances of this script to rerun, causing alpha matrix calculation to be redone
        # in each child process. Therefore, fork, which is legacy mp behavior on unix, is used.
        if sys.platform != "win32":
            ctx = mp.get_context("fork")
            with ctx.Pool(processes = cores) as pool:
                sampler = emcee.EnsembleSampler(nwalkers, ndim, mcmc_class.log_probability, pool=pool)
                sampler.run_mcmc(pos, nsteps, progress=True, store=True)

            # with Pool() as pool: # Original code
            #         sampler = emcee.EnsembleSampler(nwalkers, ndim, mcmc_class.log_probability, pool=pool)
            #         sampler.run_mcmc(pos, nsteps, progress=True)
        
        else: # Untested. Now tested, this doesn't work, needs serious modifications to make work
            with Pool() as pool:
                sampler = emcee.EnsembleSampler(nwalkers, ndim, mcmc_class.log_probability, pool=pool)
                sampler.run_mcmc(pos, nsteps, progress=verbose)


    else:
        sampler = emcee.EnsembleSampler(nwalkers, ndim, mcmc_class.log_probability)
        sampler.run_mcmc(pos, nsteps, progress=True)

    print('MCMC run takes: %s'%(time.time()-t5))

    ## discarding all vales except the last 1000 steps.
    dis_no = int(np.floor(nsteps-1000))

    global flat_samples
    ## combining all walkers together
    flat_samples = sampler.get_chain(discard=dis_no, flat=True)

    ## getting the final profile and continuum values - median of last 1000 steps
    profile = []
    global poly_cos
    poly_cos = []
    profile_err = []
    poly_cos_err = []
    
    for i in range(ndim):
        mcmc = np.median(flat_samples[:, i])
        error = np.std(flat_samples[:, i])
        mcmc = np.percentile(flat_samples[:, i], [16, 50, 84])
        error = np.diff(mcmc)
        if i<k_max:
            profile.append(mcmc[1])
            profile_err.append(np.max(error))
        else:
            poly_cos.append(mcmc[1])
            poly_cos_err.append(np.max(error))

    profile = np.array(profile)
    profile_err = np.array(profile_err)

    fig_opt = 'n'
    if fig_opt =='y':

        # plots random models from flat_samples - lets you see if it's converging
        plt.figure()
        inds = np.random.randint(len(flat_samples), size=100)
        for ind in inds:
            sample = flat_samples[ind]
            mdl = mcmc_utils.model_func(sample, x, k_max=k_max, alpha=alpha)
            mdl1 = 0
            for i in np.arange(k_max, len(sample)-1):
                mdl1 = mdl1+sample[i]*((a*x)+b)**(i-k_max)
            mdl1 = mdl1*sample[-1]
            plt.plot(x, mdl1, "C1", alpha=0.1)
            plt.plot(x, mdl, "g", alpha=0.1)
        plt.scatter(x, y, color = 'k', marker = '.', label = 'data')
        plt.xlabel("wavelengths")
        plt.ylabel("flux")
        plt.title('mcmc models and data')
        plt.savefig('figures/mcmc_and_data.png')

        prof_flux = np.exp(profile)-1

        # plots the mcmc profile - will have extra panel if it's for data
        fig, ax0 = plt.subplots()
        ax0.plot(velocities, profile, color = 'r', label = 'mcmc')
        zero_line = [0]*len(velocities)
        ax0.plot(velocities, zero_line)
        ax0.plot(velocities, model_inputs[:k_max], label = 'initial')
        ax0.fill_between(velocities, profile-profile_err, profile+profile_err, alpha = 0.3, color = 'r')
        ax0.set_xlabel('velocities')
        ax0.set_ylabel('optical depth')
        secax = ax0.secondary_yaxis('right', functions = (utils.od2flux, utils.flux2od))
        secax.set_ylabel('flux')
        ax0.legend()
        plt.savefig('figures/profile_%s'%(run_name))

        # plots mcmc continuum fit on top of data
        plt.figure('continuum fit from mcmc')
        plt.plot(x, y, color = 'k', label = 'data')
        mdl1 =0
        for i in np.arange(0, len(poly_cos)-1):
            mdl1 = mdl1+poly_cos[i]*((a*x)+b)**(i)
        mdl1 = mdl1*poly_cos[-1]
        plt.plot(x, mdl1, label = 'mcmc continuum fit')
        mdl1_poserr =0
        for i in np.arange(0, len(poly_cos)-1):
            mdl1_poserr = mdl1_poserr+(poly_cos[i]+poly_cos_err[i])*((a*x)+b)**(i)
        mdl1_poserr = mdl1_poserr*poly_cos[-1]
        mdl1_neg =0
        for i in np.arange(0, len(poly_cos)-1):
            mdl1_neg = mdl1_neg+(poly_cos[i]-poly_cos_err[i])*((a*x)+b)**(i)
        mdl1_neg = mdl1_neg*poly_cos[-1]
        plt.fill_between(x, mdl1_neg, mdl1_poserr, alpha = 0.3)
        mdl1_err =abs(mdl1-mdl1_neg)
        plt.legend()
        plt.title('continuum from mcmc')
        plt.xlabel("wavelengths")
        plt.ylabel("flux")
        plt.savefig('figures/cont_%s'%(run_name))
    
        mcmc_inputs = np.concatenate((profile, poly_cos))
        mcmc_mdl = mcmc_utils.model_func(mcmc_inputs, x, k_max=k_max, alpha=alpha)

        residuals_2 = (y+1) - (mcmc_mdl+1)

        fig, ax = plt.subplots(2,figsize=(16,9), gridspec_kw={'height_ratios': [2, 1]}, num = 'MCMC and true model', sharex = True)
        non_masked = tuple([yerr<10])
        #ax[0].plot(x, y+1, color = 'r', alpha = 0.3, label = 'data')
        #ax[0].plot(x[non_masked], mcmc_mdl[non_masked]+1, color = 'k', alpha = 0.3, label = 'mcmc spec')
        ax[1].scatter(x[non_masked], residuals_2[non_masked], marker = '.')
        ax[0].plot(x, y, 'r', alpha = 0.3, label = 'data')
        ax[0].plot(x, mcmc_mdl, 'k', alpha =0.3, label = 'mcmc spec')
        residual_masks = tuple([yerr>=100000000000000])

        #residual_masks = tuple([yerr>10])
        ax[0].scatter(x[residual_masks], y[residual_masks], label = 'masked', color = 'b', alpha = 0.3)
        ax[0].legend(loc = 'lower right')
        #ax[0].set_ylim(0, 1)
        #plotdepths = -np.array(line_depths)
        #ax[0].vlines(line_waves, plotdepths, 1, label = 'line list', color = 'c', alpha = 0.5)
        ax[1].plot(x, residuals_2, '.')
        #ax[1].scatter(x[residual_masks], residuals_2[residual_masks], label = 'masked', color = 'b', alpha = 0.3)
        z_line = [0]*len(x)
        ax[1].plot(x, z_line, '--')
        plt.savefig('figures/forward_%s'%(run_name))
        

        fig, ax0 = plt.subplots()
        ax0.plot(velocities, profile, color = 'r', label = 'mcmc')
        zero_line = [0]*len(velocities)
        ax0.plot(velocities, zero_line)
        ax0.plot(velocities, model_inputs[:k_max], label = 'initial')
        ax0.fill_between(velocities, profile-profile_err, profile+profile_err, alpha = 0.3, color = 'r')
        ax0.set_xlabel('velocities')
        ax0.set_ylabel('optical depth')
        ax0.legend()
        plt.savefig('figures/final_profile_%s'%(run_name))

    print('Getting the final profiles...')

    # finding error for the continuuum fit
    inds = np.random.randint(len(flat_samples), size=50)
    conts = []
    for ind in inds:
        sample = flat_samples[ind]
        mdl = mcmc_utils.model_func(sample, wavelengths, k_max=k_max, alpha=alpha)
        #mdl = model_func(sample, x)
        #mdl = mdl[idx]
        mdl1_temp = 0
        for i in np.arange(k_max, len(sample)-1):
            mdl1_temp = mdl1_temp+sample[i]*((a*wavelengths)+b)**(i-k_max)
        mdl1_temp = mdl1_temp*sample[-1]
        conts.append(mdl1_temp)

    continuum_error = np.std(np.array(conts), axis = 0)

    task_part = partial(get_profiles, all_frames, order, poly_cos, continuum_error)
    if len(frames)>1:
        # ctx = mp.get_context("fork") # Changed from spawn to fork to avoid multiple reruns of page
        # with ctx.Pool(processes = mp.cpu_count()) as pool:
        #     results=[pool.map(task_part, np.arange(len(frames)))]
        # results = np.array(results[0])
        # for i in range(len(frames)):
        #     all_frames[i]=results[i][i]
        for counter in range(len(frames)):
            all_frames = get_profiles(all_frames, order, poly_cos, continuum_error, counter)  
    else:
        all_frames = get_profiles(all_frames, order, poly_cos, continuum_error, 0)

    return all_frames

def ACID_HARPS(filelist, line, vgrid, order_range=None, save_path = './', file_type = 'e2ds', **kwargs):
    """Accurate Continuum fItting and Deconvolution for HARPS e2ds and s1d spectra (DRS pipeline 3.5)

    Fits the continuum of the given spectra and performs LSD on the continuum corrected spectra,
    returning an LSD profile for each file given. Files must all be kept in the same folder as well
    as their corresponding blaze files. If 's1d' are being used their e2ds equivalents must also be
    in this folder. Result files containing profiles and associated errors for each order (or
    corresponding wavelength range in the case of 's1d' files) will be created and saved to a
    specified folder. It is recommended that this folder is seperate to the input files.

    Parameters
    ----------
    filelist : list of strings
        List of files. Files must come from the same observation night as continuum is fit for a combined
        spectrum of all frames. A profile and associated errors will be produced for each file specified.
    line : str
        Path to linelist. Takes VALD linelist in long or short format as input. Minimum line depth input into VALD must
        be less than 1/(3*SN) where SN is the highest signal-to-noise ratio of the spectra. 
    vgrid : array
        Velocity grid for LSD profiles (in km/s).
    order_range : array, optional
        Orders to be included in the final profiles. If s1d files are input, the corresponding wavelengths 
        will be considered, by default None, the function generates np.arange(10,70) in the body.
    save_path : str, optional
        Path to the directory where output files will be saved, by default './'
    file_type : str, optional
        Type of the input files, either "e2ds" or "s1d", by default 'e2ds'
    **kwargs
        Additional arguments to be passed to the ACID function. See ACID function for details.

    Returns
    -------
    list
        Barycentric Julian Date for files
    list
        Profiles (in normalised flux)
    list
        Errors on profiles (in normalised flux)
    """

    global velocities
    velocities = vgrid.copy()
    global all_frames
    all_frames = np.zeros((len(filelist), len(order_range), 2, len(velocities)))
    global linelist
    linelist = line

    global frames
    global frame_wavelengths
    global frame_errors
    global sns

    if order_range is None:
        order_range = np.arange(10, 70)
    
    for order in order_range:
    
        print('Running for order %s/%s...'%(order-min(order_range)+1, max(order_range)-min(order_range)+1))

        frame_wavelengths, frames, frame_errors, sns, telluric_spec = read_in_frames(order, filelist, file_type)

        all_frames = ACID(frame_wavelengths, frames, frame_errors, linelist, sns, velocities, all_frames, order=order-min(order_range), **kwargs)

    # adding into fits files for each frame
    BJDs = []
    profiles = []
    errors = []
    for frame_no in range(0, len(frames)):
        file = filelist[frame_no]
        fits_file = fits.open(file)
        hdu = fits.HDUList()
        hdr = fits.Header()
        
        for order in order_range:
            hdr['ORDER'] = order
            hdr['BJD'] = fits_file[0].header['ESO DRS BJD']
            if order == order_range[0]:
                BJDs.append(fits_file[0].header['ESO DRS BJD'])
            hdr['CRVAL1']=np.min(velocities)
            hdr['CDELT1']=velocities[1]-velocities[0]

            profile = all_frames[frame_no, order-min(order_range), 0]
            profile_err = all_frames[frame_no, order-min(order_range), 1]

            hdu.append(fits.PrimaryHDU(data = [profile, profile_err], header = hdr))
            if save_path != 'no save':
                month = 'August2007'
                hdu.writeto('%s%s_%s_%s.fits'%(save_path, month, frame_no, run_name), output_verify = 'fix', overwrite = 'True')

        result1, result2 = combineprofiles(all_frames[frame_no, :, 0], all_frames[frame_no, :, 1])
        profiles.append(result1)
        errors.append(result2)

    return BJDs, profiles, errors



import numpy as np
from scipy import linalg
from astropy.io import  fits
import glob, time, warnings, sys, psutil, os
import scipy.constants as const
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from scipy.interpolate import interp1d, LSQUnivariateSpline
from tqdm import tqdm
ckms = float(const.c/1e3)  # speed of light in km/s

def LSD(wavelengths, flux_obs, rms, linelist, adjust_continuum, poly_ord, sn, 
        order, run_name, velocities, verbose=False):

    t0 = time.time()

    #idx = tuple([flux_obs>0])
    # Converting to optical depth space
    rms = rms / flux_obs
    flux_obs = np.log(flux_obs)

    deltav = velocities[1] - velocities[0]

    #### This is the EXPECTED linelist (for a slow rotator of the same spectral type) ####
    # Ben - Reading the linelist
    linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
    wavelengths_expected_all = np.array(linelist_expected[:,0])
    depths_expected_all = np.array(linelist_expected[:,1])

    # Selecting lines within the wavelength range of the observed spectrum
    wavelength_min = np.min(wavelengths)
    wavelength_max = np.max(wavelengths)
    idx = np.logical_and(wavelengths_expected_all >= wavelength_min,
                         wavelengths_expected_all <= wavelength_max)
    wavelengths_expected = wavelengths_expected_all[idx]
    depths_expected = depths_expected_all[idx]

    # Selecting lines deeper than 1/(3*sn)
    line_min = 1 / (3 * sn)
    idx = (depths_expected >= line_min)
    wavelengths_expected = wavelengths_expected[idx]
    depths_expected = depths_expected[idx]
    no_line = len(depths_expected)
    
    # Converting to log space for depths
    depths_expected = -np.log(1-depths_expected)

    blankwaves = wavelengths
    R_matrix = flux_obs

    # Find differences and velocities
    diff = blankwaves[:, np.newaxis] - wavelengths_expected
    vel = ckms * (diff / wavelengths_expected)

    # We can calculate the alpha matrix in one pass if the number of wavelengths is small enough
    
    # Note this used to be "if len(wavelengths)<6000", which I believe is way too high
    # for 16GB RAM. With testing, I found that if the matrix was half the available memory,
    # it would always be fast, otherwise seperate into blocks as the else statment below.
    m_size = len(wavelengths_expected) * len(velocities) * len(wavelengths) * 8 * 1e-9 # Matrix size in GB
    m_available = psutil.virtual_memory().available * 1e-9
    if m_size < m_available / 2:
        
        x = (vel[:, :, np.newaxis] - velocities) / deltav
        delta = np.clip(1.0 - np.abs(x), 0.0, 1.0)
        alpha = (depths_expected[:, None] * delta).sum(axis=1)  # (nb, n_vel)

        # The below was simplified using np.clip (as shown above)
        # alpha_mask_1 = np.logical_and(-1. < x, x < 0.)
        # alpha_mask_2 = np.logical_and(0. <= x, x < 1.)
        # delta_x_1 = 1 + x
        # delta_x_2 = 1 - x
        # delta_x_1[alpha_mask_1==False] = 0
        # delta_x_2[alpha_mask_2==False] = 0
        # # Update alpha array using calculated delta_x values
        # alpha = np.zeros((len(blankwaves), len(velocities)))
        # alpha += (depths_expected[:, np.newaxis] * delta_x_1).sum(axis=1)
        # alpha += (depths_expected[:, np.newaxis] * delta_x_2).sum(axis=1)

    else:
        block = 512 # after initial testing, this value is a good compromise between memory use and speed
        alpha  = np.zeros((len(blankwaves), len(velocities)), dtype=np.float64)
        if verbose:
            iterable = tqdm(range(0, len(wavelengths_expected), block), desc='Calculating alpha matrix')
        else:
            iterable = range(0, len(wavelengths_expected), block)
        for start_pos in iterable:
            end_pos = min(start_pos + block, len(wavelengths_expected)) # ensure we don't go out of bounds on last iteration
            wl  = wavelengths_expected[start_pos:end_pos]
            dep = depths_expected[start_pos:end_pos]

            vel_blk = ckms * (blankwaves[:, None] - wl) / wl
            x_blk   = (vel_blk[:, :, None] - velocities) / deltav
            delta   = np.clip(1.0 - np.abs(x_blk), 0.0, 1.0)                    

            alpha += (dep[:, None] * delta).sum(axis=1)

    # The below I think works similarly to above, but but uses only for loops, rather then using a for loop
    # that only involves numpy calculations. With testing, the above is much faster.
    # else:
    #     warnings.warn('Large wavelength ranges give large computation time. Seperate wavelength range into smaller chunks for faster computation.', DeprecationWarning, stacklevel=2)
    #     alpha = np.zeros((len(blankwaves), len(velocities)))

    #     for j in tqdm(range(0, len(blankwaves)), desc='Calculating alpha matrix'):
    #         for i in (range(0, len(wavelengths_expected))):
    #             vdiff = ((blankwaves[j] - wavelengths_expected[i]) * ckms) / wavelengths_expected[i]
    #             if vdiff <= (np.max(velocities) + deltav) and vdiff >= (np.min(velocities) - deltav):
    #                 diff = blankwaves[j] - wavelengths_expected[i]
    #                 vel = const.c / 1e3 * (diff / wavelengths_expected[i])
    #                 for k in range(0, len(velocities)):
    #                     x = (velocities[k] - vel) / deltav
    #                     if -1. < x and x < 0.:
    #                         delta_x = (1 + x)
    #                         alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
    #                     elif 0. <= x and x < 1.:
    #                         delta_x = (1 - x)
    #                         alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
    #             else:
    #                 pass

    id_matrix = np.identity(len(flux_obs))
    S_matrix = (1/rms) * id_matrix

    S_squared = np.dot(S_matrix, S_matrix)
    alpha_transpose = (np.transpose(alpha))

    RHS_1 = np.dot(alpha_transpose, S_squared)
    RHS_final = np.dot(RHS_1, R_matrix)

    LHS_preinvert = np.dot(RHS_1, alpha)
    LHS_prep = np.matrix(LHS_preinvert)

    P, L, U = linalg.lu(LHS_prep)

    n = len(LHS_prep)
    B = np.identity(n)
    Z = linalg.solve_triangular(L, B, lower=True)
    X = linalg.solve_triangular(U, Z, lower=False)
    LHS_final = np.matmul(X, np.transpose(P))

    profile = np.dot(LHS_final, RHS_final)
    profile_errors_squared = np.diagonal(LHS_final)
    profile_errors = np.sqrt(profile_errors_squared)

    return velocities, profile, profile_errors, alpha, wavelengths_expected, depths_expected_all, no_line

def get_wave(data,header):

  wave=np.array(data*0., dtype = 'longdouble')
  no=data.shape[0]
  npix=data.shape[1]
  d=header['ESO DRS CAL TH DEG LL']
  xx0=np.arange(npix)
  xx=[]
  for i in range(d+1):
      xx.append(xx0**i)
  xx=np.asarray(xx, dtype = 'longdouble')

  for o in range(no):
      for i in range(d+1):
          idx=i+o*(d+1)
          par=np.longdouble(header['ESO DRS CAL TH COEFF LL%d' % idx])
          wave[o,:]=wave[o,:]+par*xx[i,:]
       #for x in range(npix):
       #  wave[o,x]=wave[o,x]+par*xx[i,x]#float(x)**float(i)

  return wave

def continuumfit(wavelengths1, fluxes1, poly_ord):

        fluxes = fluxes1
        wavelengths = wavelengths1

        idx = wavelengths.argsort()
        wavelength = wavelengths[idx]
        fluxe = fluxes[idx]
        clipped_flux = []
        clipped_waves = []
        binsize = 100
        for i in range(0, len(wavelength), binsize):
            waves = wavelength[i:i+binsize]
            flux = fluxe[i:i+binsize]
            indicies = flux.argsort()
            flux = flux[indicies]
            waves = waves[indicies]

            clipped_flux.append(flux[len(flux)-1])
            clipped_waves.append(waves[len(waves)-1])
        coeffs=np.polyfit(clipped_waves, clipped_flux, poly_ord)

        poly = np.poly1d(coeffs)
        fit = poly(wavelengths1)
        # plt.figure()
        # plt.plot(wavelengths1, fluxes1)
        # plt.plot(wavelengths1, fit)
    
        flux_obs = fluxes1/fit
        
        return flux_obs

def upper_envelope(x, y):
    # from MM-LSD code - give credit if needed
    # used to compute the tapas continuum. find peaks then fit spline to it.
    peaks = find_peaks(y, height=0.2, distance=len(x) // 500)[0]
    # t= knot positions
    spl = LSQUnivariateSpline(x=x[peaks], y=y[peaks], t=x[peaks][5::10])
    return spl(x)

def blaze_correct(file_type, spec_type, order, file, directory, masking, run_name, berv_opt):
    #### Inputing spectrum depending on file_type and spec_type #####

    if file_type == 's1d':
        #### Finding min and max wavelength from e2ds for the specified order ######
        file_e2ds = file.replace('s1d', 'e2ds')
        # print(file_e2ds)
        hdu=fits.open('%s'%file_e2ds)
        sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]
        spec=hdu[0].data
        header=hdu[0].header
        brv=header['ESO DRS BERV']
        # print('hi')
        spec_check = spec[spec<=0]
        # if len(spec_check)>0:
            # print('WARNING NEGATIVE/ZERO FLUX - corrected')

        where_are_zeros = (spec<=0)
        spec[where_are_zeros] = 1000000000000
        flux_error = np.sqrt(spec)

        wave=get_wave(spec, header)*(1.+brv/2.99792458e5)
        wavelengths_order = wave[order]
        wavelength_min = np.min(wavelengths_order)
        wavelength_max = np.max(wavelengths_order)

        ## remove overlapping region (always remove the overlap at the start of the order, i.e the min_overlap)
        last_wavelengths = wave[order-1]
        next_wavelengths = wave[order+1]
        min_overlap = np.max(last_wavelengths)
        max_overlap = np.min(next_wavelengths)

        # print(min_overlap)
        # print(max_overlap)
        # plt.figure()
        # plt.plot(wave[order-1], spec[order-1])
        # plt.plot(wave[order], spec[order])
        # plt.plot(wave[order+1], spec[order+1])
        # plt.show()

        # idx_ = tuple([wavelengths>min_overlap])
        # wavelength_min = 5900
        # wavelength_max = wavelength_min+200    ###### if you want to do a WAVELENGTH RANGE just input min and max here ######
        #print(wavelength_max)
        hdu.close()

        # plt.figure('e2ds vs s1d, straight from fits')
        # plt.plot(wavelengths_order, spec[order], label = 'e2ds')

        #### Now reading in s1d file ########
        # print(file)
        hdu=fits.open('%s'%file)
        spec=hdu[0].data
        header=hdu[0].header
        spec_check = spec[spec<=0]

        # print(hdu[0].header['CRVAL1'])
        # print(hdu[0].header['CRPIX1'])
        # print(hdu[0].header['CDELT1'])
        # print(spec.shape[0])
        # print((hdu[0].header['CRPIX1']+np.arange(spec.shape[0])))
        # print((np.arange(spec.shape[0])))
        # print((hdu[0].header['CRPIX1']+np.arange(spec.shape[0]))*hdu[0].header['CDELT1'])
        # print((np.arange(spec.shape[0]))*hdu[0].header['CDELT1'])

        # wave=hdu[0].header['CRVAL1']+(hdu[0].header['CRPIX1']+np.arange(spec.shape[0]))*hdu[0].header['CDELT1']
        
        wave=hdu[0].header['CRVAL1']+(np.arange(spec.shape[0]))*hdu[0].header['CDELT1']
        
        where_are_zeros = (spec<=0)
        spec[where_are_zeros] = 1000000000000
        flux_error = np.sqrt(spec)

        # print(wave)

        if spec_type == 'order':
            wavelengths=[]
            fluxes=[]
            errors = []
            for value in range(0, len(wave)):
                l_wave=wave[value]
                l_flux=spec[value]
                l_error=flux_error[value]
                if l_wave>=wavelength_min and l_wave<=wavelength_max:
                    wavelengths.append(l_wave)
                    fluxes.append(l_flux)
                    errors.append(l_error)
            wavelengths = np.array(wavelengths)
            fluxes = np.array(fluxes)
            errors = np.array(errors)


            if len(wavelengths)>5144:
                wavelengths = wavelengths[:5144]
                fluxes = fluxes[:5144]
                flux_error = errors[:5144]
            # print(len(wavelengths))
            # print(np.max(wavelengths), np.min(wavelengths))
            # print(min_overlap)
            # print(max_overlap)
            # idx_overlap = np.logical_and(wavelengths>=min_overlap, wavelengths<=max_overlap)
            # idx_overlap = tuple([idx_overlap==False])
            overlap = []

            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 'Flux')
            # # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
            # plt.show()

            # plt.plot(wavelengths, fluxes, label = 's1d')
            # plt.legend()
            # plt.show()

            spec_check = fluxes[fluxes<=0]
            # if len(spec_check)>0:
            #     print('WARNING NEGATIVE/ZERO FLUX - corrected')

            # where_are_zeros = (spec<=0)
            # spec[where_are_zeros] = 1000000000000
            # flux_error = np.sqrt(spec)
            
            flux_error_order = flux_error
            masked_waves = []
            masked_waves = np.array(masked_waves)

            # ## test - s1d interpolated onto e2ds wavelength grid ##
            # hdu_e2ds=fits.open('%s'%file.replace('s1d', 'e2ds'))
            # spec_e2ds=hdu_e2ds[0].data
            # header_e2ds=hdu_e2ds[0].header

            # wave_e2ds=get_wave(spec_e2ds, header_e2ds)*(1.+brv/2.99792458e5)

            # # plt.figure()
            # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), wave_e2ds[order][1:]-wave_e2ds[order][:-1], label = 'e2ds wave (after berv)')
            # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), get_wave(spec_e2ds, header_e2ds)[order][1:]-get_wave(spec_e2ds, header_e2ds)[order][:-1], label = 'e2ds wave (before berv)')
            # # # plt.scatter(np.arange(len(wavelengths[:-1])), wavelengths[:-1]-wavelengths[1:], label = 's1d wave')
            # # plt.legend()
            # # plt.show()

            # # id = np.logical_and(wave_e2ds<np.max(wavelengths), wave_e2ds>np.min(wavelengths))
            # # print(wave_e2ds*u.AA)
            # # print(wavelengths*u.AA)
            # # print(fluxes*u.photon)

            # blaze_file = glob.glob('%sblaze_folder/**blaze_A*.fits'%(directory))
            # # print('%sblaze_folder/**blaze_A*.fits'%(directory))
            # # print(blaze_file)
            # blaze_file = blaze_file[0]
            # blaze =fits.open('%s'%blaze_file)
            # blaze_func = blaze[0].data
            # spec_e2ds = spec_e2ds/blaze_func
        
            # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # print(diff_arr)
            # wavelengths = wavelengths[:-1]
            # fluxes = fluxes[:-1]/diff_arr

            # s1d_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
            # fluxcon = FluxConservingResampler()
            # new_spec = fluxcon(s1d_spec, wave_e2ds[order]*u.AA)

            # wavelengths = new_spec.spectral_axis
            # fluxes = new_spec.flux

            # wavelengths = wavelengths[10:len(wave_e2ds[order])-9]/u.AA
            # fluxes = fluxes[10:len(wave_e2ds[order])-9]/u.Unit('photon AA-1')
            # flux_error_order = flux_error_order[10:len(wave_e2ds[order])-10]

            # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # print(diff_arr)
            # wavelengths = wavelengths[:-1]
            # fluxes = fluxes[:-1]*diff_arr

            # print(wavelengths)
            # print(fluxes)

            # plt.figure()
            # plt.title('interpolated s1d comapred to actual e2ds spectrum')
            # plt.plot(wavelengths, fluxes, label = 'interpolated s1d on e2ds wave grid')
            # plt.plot(wave_e2ds[order], spec_e2ds[order], label = 'e2ds spectrum')
            # plt.legend()
            # plt.show()


            # ## end of test ##

            # ## test2 - synthetic s1d spectrum - using wavelength grid ##
            # def gauss(x1, rv, sd, height, cont):
            #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
            #     return y1

            # wavelength_grid = wavelengths
            # flux_grid = np.ones(wavelength_grid.shape)

            # linelist = '/home/lsd/Documents/fulllinelist0001.txt'
            
            # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
            # wavelengths_expected1 =np.array(linelist_expected[:,0])
            # depths_expected1 = np.array(linelist_expected[:,1])
            # # print(len(depths_expected1))

            # wavelength_min = np.min(wavelengths)
            # wavelength_max = np.max(wavelengths)

            # print(wavelength_min, wavelength_max)

            # wavelengths_expected=[]
            # depths_expected=[]
            # no_line =[]
            # for some in range(0, len(wavelengths_expected1)):
            #     line_min = 0.25
            #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
            #         wavelengths_expected.append(wavelengths_expected1[some])
            #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
            #         depths_expected.append(depths_expected1[some])
            #     else:
            #         pass
            
            # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
            # count_range = np.array(count_range, dtype = int)
            # print(count_range)
            # vgrid = np.linspace(-21,18,48)
            # try: ccf = fits.open(file.replace('s1d', 'ccf_K5'))
            # except: ccf = fits.open(file.replace('s1d', 'ccf_G2'))
            # rv = ccf[0].header['ESO DRS CCF RV']

            
            # for line in count_range:
            #     mid_wave = wavelengths_expected[line]
            #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
            #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
            #     prof_wavelength_grid = wavelength_grid[id]
            #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
            #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
            #     # plt.figure()
            #     # plt.plot(prof_wavelength_grid, prof)
            #     # plt.show()
            #     flux_grid[id] = prof

            # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
            # poly = np.poly1d(coeffs*fluxes[0])
            # fit = poly(wavelengths)

            # wavelengths = wavelength_grid
            # fluxes = flux_grid * fit

            if masking == 'masked':
                ## I've just been updating as I go through so it's not complete
                #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
                masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
                min_waves_mask = np.array(masks_csv[:,0])
                max_waves_mask = np.array(masks_csv[:,1])
                masks = []
                for mask_no in range(len(min_waves_mask)):
                    masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

                masked_waves=[]

                for mask in masks:
                    #print(np.max(mask), np.min(mask))
                    idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                    #print(flux_error_order[idx])
                    flux_error_order[idx] = 10000000000000000000
                    #print(idx)
                    if len(wavelengths[idx])>0:
                        masked_waves.append(wavelengths[idx])

                #masks = []
                ### allows extra masking to be added ##

                plt.figure('masking')
                plt.plot(wavelengths, fluxes)
                if len(masked_waves)>0:
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                #plt.show()

                #response = input('Are there any regions to be masked? y or n: ')
                response = 'y'
                if response == 'y':
                    '''
                    print('Take note of regions.')
                    plt.figure('masking')
                    plt.plot(wavelengths, fluxes)
                    plt.show()
                    '''
                    #response1 = int(input('How many regions to mask?: '))
                    response1 = 0
                    for i in range(response1):
                        min_wave = float(input('Minimum wavelength of region %s: '%i))
                        max_wave = float(input('Maximum wavelength of region %s: '%i))
                        masks.append([min_wave, max_wave])
                    masked_waves=[]
                    #print(masks)
                    for mask in masks:
                        print(np.max(mask), np.min(mask))
                        idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                        #print(flux_error_order[idx])
                        flux_error_order[idx] = 10000000000000000000

                        if len(wavelengths[idx])>0:
                            masked_waves.append(wavelengths[idx])

                    plt.figure('telluric masking')
                    plt.title('Spectrum - after telluric masking')
                    plt.plot(wavelengths, fluxes)
                    # print(masked_waves)
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                    #print('new version')
                    plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))

                    plt.figure('errors')
                    plt.plot(wavelengths, flux_error_order)
                    plt.close('all')
                    #plt.show()

                if response == 'n':
                    print('yay!')


            elif masking == 'unmasked':
                masked_waves = []
                masked_waves = np.array(masked_waves)

        elif spec_type == 'full':
            ## not set up properly.
            wavelengths = wave
            fluxes = spec

    elif file_type == 'e2ds':
        hdu=fits.open('%s'%file)
        spec=hdu[0].data
        header=hdu[0].header
        sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]
        # print('S/N: %s'%sn)
        spec_check = spec[spec<=0]
        # if len(spec_check)>0:
        #     print('WARNING NEGATIVE/ZERO FLUX - corrected')


        # where_are_NaNs = np.isnan(flux_error)
        # flux_error[where_are_NaNs] = 1000000000000
        where_are_zeros = (spec<=0)
        spec[where_are_zeros] = 1000000000000
        flux_error = np.sqrt(spec)
        '''
        flux_error1 = header['HIERARCH ESO DRS SPE EXT SN%s'%order]

        flux_error = header['HIERARCH ESO DRS CAL TH RMS ORDER%s'%order]
        print(flux_error, flux_error1)

        flux_error = flux_error*np.ones(np.shape(spec))
        '''
        # file_ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
        # print(file_ccf[0].header['ESO DRS BERV'])
        brv=np.longdouble(header['ESO DRS BERV'])
        # print(brv)
        wave_nonad=get_wave(spec, header)
        # if berv_opt == 'y':
        #     print('BERV corrected')
        wave = wave_nonad*(1.+brv/2.99792458e5)
        wave = np.array(wave, dtype = 'double')
        # if berv_opt == 'n':
        #     print('BERV not corrected')
        # wave = wave_nonad
        
        # plt.figure('before blaze correction - e2ds vs s1d - after berv correction')
        # plt.plot(wave[order], spec[order], label = 'e2ds')
        # plt.plot(wave_s1d[wave_s1d<np.max(wave[order])], spec_s1d[wave_s1d<np.max(wave[order])], label = 's1d')
        # plt.show()

        # plt.figure('before and after berv')
        # plt.plot(wave_nonad[order], spec[order], label = 'before berv correction')
        # plt.plot(wave[order], spec[order], label = 'after berv correction')
        # plt.show()
    
        # rv_drift=header['ESO DRS DRIFT RV'] 
        # print(rv_drift)
        wave_corr = (1.+brv/2.99792458e5)
        # print(brv, (wave_corr-1)*2.99792458e5)

        # inp = input('Enter to continue...')
        '''
        plt.figure('Spectrum directly from fits file')
        plt.title('Spectrum directly from fits file')
        plt.errorbar(wave[order], spec[order], yerr = flux_error[order])
        plt.xlabel('wavelengths')
        plt.ylabel('flux')
        plt.show()
        '''

        blaze_file = glob.glob('tests/data/*blaze_B*.fits')

        blaze_file = blaze_file[0]
        # try:
        #     blaze_file = glob.glob('./**blaze_A*.fits')
        #     # print('%sblaze_folder/**blaze_A*.fits'%(directory))
        #     # print(blaze_file)
        #     blaze_file = blaze_file[0]
        # except: 
        #     try:
        #         blaze_file = glob.glob('/Users/lucydolan/Starbase/problem_frames/**blaze_A*.fits')
        #         blaze_file = blaze_file[0]
        #     except:
        #         blaze_file = glob.glob('tests/data/**blaze_A*.fits')
        blaze = fits.open('%s'%blaze_file)
        blaze_func = blaze[0].data
        min_rows = min(spec.shape[0], blaze_func.shape[0], flux_error.shape[0])
        spec, blaze_func, flux_error = spec[:min_rows], blaze_func[:min_rows], flux_error[:min_rows]
        spec = spec/blaze_func
        flux_error = flux_error/blaze_func

        # plt.figure()
        # plt.figure('blaze for orders 28, 29 and 30')
        # plt.plot(wave[28], blaze[0].data[28])
        # plt.plot(wave[29], blaze[0].data[29])
        # plt.plot(wave[30], blaze[0].data[30])

        # plt.figure()
        # plt.title('blaze for orders 28, 29 and 30 summed together')
        # pixel_grid = np.linspace(np.min(wave[28]), np.max(wave[30]), len(np.unique(wave[28:30])))
        # blaze_sum = np.zeros(pixel_grid.shape)

        # f28 = interp1d(wave[28], blaze[0].data[28], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
        # f29 = interp1d(wave[29], blaze[0].data[29], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
        # f30 = interp1d(wave[30], blaze[0].data[30], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')

        # idx28 = np.logical_and(pixel_grid>np.min(wave[28]), pixel_grid<np.max(wave[28]))
        # idx29 = np.logical_and(pixel_grid>np.min(wave[29]), pixel_grid<np.max(wave[29]))
        # idx30 = np.logical_and(pixel_grid>np.min(wave[30]), pixel_grid<np.max(wave[30]))

        # blaze_28 = f28(pixel_grid[idx28])
        # blaze_29 = f29(pixel_grid[idx29])
        # blaze_30 = f30(pixel_grid[idx30])

        # wave28 = pixel_grid[idx28]
        # wave29 = pixel_grid[idx29]
        # wave30 = pixel_grid[idx30]

        # for pixel in range(len(pixel_grid)):
        #     wavell = pixel_grid[pixel]
            
        #     idx28 = tuple([wave28==wavell])
        #     idx29 = tuple([wave29==wavell])
        #     idx30 = tuple([wave30==wavell])
            
        #     try: b = blaze_28[idx28][0]
        #     except: b=0
        #     try: b1 = blaze_29[idx29][0]
        #     except: b1=0
        #     try: b2 = blaze_30[idx30][0]
        #     except: b2=0

        #     print(wavell, b, b1, b2)

        #     blaze_sum[pixel] = b + b1 + b2 

        # plt.plot(pixel_grid, blaze_sum)
        # plt.show()

        # plt.figure()
        # plt.title('e2ds after blaze orders 28, 29 and 30')
        # plt.plot(wave[28], spec[28])
        # plt.plot(wave[29], spec[29])
        # plt.plot(wave[30], spec[30])
        
        ## TEST - adjusting e2ds spectrum onto s1d continuum ##

        # ## first interpolate s1d onto e2ds wavelength grid ##
        # s1d_file = fits.open(file.replace('e2ds', 's1d'))
        # s1d_spec = s1d_file[0].data
        # wave_s1d = s1d_file[0].header['CRVAL1']+(np.arange(s1d_spec.shape[0]))*s1d_file[0].header['CDELT1']

        # wavelengths = wave_s1d
        # fluxes = s1d_spec

        # plt.figure()
        # plt.plot(wavelengths, fluxes, label = 'e2ds spectrum - corrected to s1d continuum')
        # plt.plot(wave_s1d, s1d_spec, label = 's1d spectrum')
        # plt.legend()
        # plt.show()

        # plt.figure()
        # plt.plot(wavelengths, fluxes, label = 's1d')
        # plt.legend()

        # diff_arr = wavelengths[1:] - wavelengths[:-1]
        # # print(diff_arr)
        # wavelengths = wavelengths[:-1]
        # fluxes = fluxes[:-1]

        # fluxes = fluxes/diff_arr

        # fluxes = np.ones(fluxes.shape)
        # for i in range(len(fluxes)):
        #     fluxes[i] = fluxes[i]*8000
        # plt.figure()
        # plt.plot(wavelengths, fluxes, label = 's1d in photons AA-1')
        # #plt.legend()

        # interpolate s1d onto e2ds wavlengths - non flux conserving
        # s1dd_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
        # fluxcon = FluxConservingResampler()
        # extended_e2ds_wave = np.concatenate((wave[order], [wave[order][-1]+0.01]))

         ## MM-LSD way
    
        fluxes = spec[order]
        flux_error_order = flux_error[order]
        wavelengths = wave[order]

        # for i in range(len(fluxes)):
        #     if i ==0:
        #         # print(fluxes[i])
        #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[1]-wavelengths[0]))
        #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
        #         # print(fluxes[i])
        #     else:
        #         # print(fluxes[i])
        #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[i]-wavelengths[i-1]))
        #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
        #         # print(fluxes[i])

        # ## end of MM-LSD way

        # new_spec = fluxcon(s1dd_spec, wavelengths*u.AA)

        # reference_wave = new_spec.spectral_axis/u.AA
        # reference_flux = new_spec.flux/u.Unit('photon AA-1')

        # # print(len(wavelengths_new))
        # # print(len(wave[order]))

        # # # plt.plot(wavelengths_new, fluxes, label = 'interpolated s1d in photons AA-1')
        # # # plt.legend()

        # #diff_arr = wavelengths_new[1:] - wavelengths_new[:-1]
        # #reference_wave = wavelengths_new[:-1]
        # #reference_flux = fluxes[:-1]*diff_arr
        
        # # # print(len(reference_wave))
        # # # print(len(wave[order]))

        # # # print(reference_wave-wave[order])

        # # # plt.figure()
        # # # plt.plot(reference_wave, reference_flux, label = 'interpolated s1d in photons per bin')
        # # # plt.legend()
        # # # plt.show()
        # # ## divide e2ds spectrum by interpolated s1d and fit polynomial to result

        # reference_wave = np.array(reference_wave, dtype = float)
        # reference_flux = np.array(reference_flux, dtype = float)
        # div_frame = fluxes/reference_flux

        # # plt.figure()
        # # plt.plot(reference_wave, div_frame)
        # # plt.show()

        # # # ### creating windows to fit polynomial to
        # # # binned = np.zeros(int(len(div_frame)/2))
        # # # binned_waves = np.zeros(int(len(div_frame)/2))
        # # # for i in range(0, len(div_frame)-1, 2):
        # # #     pos = int(i/2)
        # # #     binned[pos] = (div_frame[i]+div_frame[i+1])/2
        # # #     binned_waves[pos] = (reference_wave[i]+reference_wave[i+1])/2

        # # # plt.plot(frame_wavelengths[n], frames_unadjusted[n], color = 'b', label = 'unadjusted')
        # # # plt.figure()
        # # # plt.plot(frame_wavelengths[n], frames[n])
        # # # plt.show()

        # ### fitting polynomial to div_frame
        # coeffs=np.polyfit(reference_wave, div_frame, 3)
        # poly = np.poly1d(coeffs)
        # # print(coeffs)
        # inputs = coeffs[::-1]
        # # print(inputs)

        # wavelengths = reference_wave

        # fit = poly(wavelengths)
       
        # # # plt.figure()
        # # # plt.plot(reference_wave, reference_flux, label= 'reference')
        # # # plt.plot(wave[order], spec[order], label = 'e2ds')
        # # # plt.legend()

        # # # plt.figure()
        # # # plt.plot(reference_wave-wavelengths[:-1], label = 'reference_wave-wavelengths')
        # # # plt.legend()
        # # # plt.show()

        # # plt.figure()
        # # plt.scatter(reference_wave, div_frame, label = 'div flux')
        # # plt.plot(wavelengths, fit, label = 'fit')
        # # # plt.plot(reference_wave, poly(reference_wave), label = 'poly(reference_wave)')
        # # plt.legend()
        # # plt.show()
        
        # fluxes = spec[order]/fit
        # flux_error_order = flux_error[order]/fit

        # # plt.figure()
        # # plt.plot(wavelengths, fluxes/reference_flux, label = 'continuum adjusted e2ds/s1d')
        # # plt.legend()
        # # plt.show()

        # # plt.figure()
        # # plt.plot(wavelengths, fluxes-reference_flux, label = 'continuum adjusted e2ds-s1d')
        # # plt.legend()
        # # plt.show()
        # # plt.figure()
        # # plt.plot(wavelengths, spec[order], label = 'before')
        # # plt.plot(wavelengths, fit)
        # # plt.plot(wavelengths, fluxes, label= 'after')
        # # plt.legend()
        # # plt.show()

        # # idx_full = np.logical_and(wave_s1d>np.min(wave[28]), wave_s1d<np.max(wave[30]))
        # # plt.plot(wave_s1d[idx_full], s1d_spec[idx_full])
        # # plt.show()

        # blaze.close()
        # # plt.figure('after blaze correction - e2ds vs s1d - after berv correction')
        # # plt.plot(wave[order], spec[order], label = 'e2ds')
        # # plt.plot(wave_s1d[wave_s1d>np.max(wave[order])], spec_s1d[wave_s1d>np.max(wave[order])], label = 's1d')
        # # plt.show()

        

        # # # test - e2ds interpolated onto s1d wavelength grid ##
        # # hdu_s1d=fits.open('%s'%file.replace('e2ds', 's1d'))
        # # spec_s1d=hdu_s1d[0].data
        # # header_s1d=hdu_s1d[0].header

        # # wave_s1d=header_s1d['CRVAL1']+(header_s1d['CRPIX1']+np.arange(spec_s1d.shape[0]))*header_s1d['CDELT1']
        # # id = np.logical_and(wave_s1d<np.max(wavelengths), wave_s1d>np.min(wavelengths))
        # # print(wave_s1d*u.AA)
        # # print(wavelengths*u.AA)
        # # print(fluxes*u.Unit('erg cm-2 s-1 AA-1'))
        # # # plt.figure('s1d compared to interpolated e2ds')
        # # # plt.title('s1d compared to interpolated e2ds')
        # # # plt.plot(wave_s1d, spec_s1d, label = 's1d spectrum')

        # # ## these fluxes are in photons per bin - I need them in photons per Angstrom
        # # ## therefore i do flux/angstroms in pixel
        # # diff_arr = wavelengths[1:] - wavelengths[:-1]
        # # print(diff_arr)
        # # wavelengths = wavelengths[:-1]
        # # fluxes = fluxes[:-1]
        # # # plt.figure('changing flux units')
        # # # plt.plot(wavelengths, fluxes, label = 'flux per pixel')

        # # fluxes = fluxes/diff_arr

        # # # plt.plot(wavelengths, fluxes, label = ' flux per A')

        # # e2ds_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
        # # fluxcon = FluxConservingResampler()
        # # new_spec = fluxcon(e2ds_spec, wave_s1d[id]*u.AA)

        # # wavelengths = new_spec.spectral_axis
        # # fluxes = new_spec.flux

        # # wavelengths = wavelengths[:4097]/u.AA
        # # fluxes = fluxes[:4097]/u.Unit('photon AA-1')

        # # diff_arr = wavelengths[1:] - wavelengths[:-1]
        # # wavelengths = wavelengths[:-1]
        # # fluxes = fluxes[:-1]*diff_arr

        # # # plt.figure('s1d compared to interpolated e2ds') 
        # # # plt.plot(wavelengths, fluxes, label = 'interpolated e2ds onto s1d')
        # # # plt.xlim(np.min(wavelengths), np.max(wavelengths))
        # # # plt.legend()
        # # # plt.show() 

        # # print(wavelengths)
        # # print(fluxes)

        # #end of test ##

        # # # test2 - synthetic e2ds spectrum - using wavelength grid ##
        # # def gauss(x1, rv, sd, height, cont):
        # #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
        # #     return y1

        # # wavelength_grid = wavelengths
        # # flux_grid = np.ones(wavelength_grid.shape)

        # # linelist = '/home/lsd/Documents/fulllinelist0001.txt'

        # # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
        # # wavelengths_expected1 =np.array(linelist_expected[:,0])
        # # depths_expected1 = np.array(linelist_expected[:,1])
        # # # print(len(depths_expected1))

        # # wavelength_min = np.min(wavelengths)
        # # wavelength_max = np.max(wavelengths)

        # # print(wavelength_min, wavelength_max)

        # # wavelengths_expected=[]
        # # depths_expected=[]
        # # no_line =[]
        # # for some in range(0, len(wavelengths_expected1)):
        # #     line_min = 0.25
        # #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
        # #         wavelengths_expected.append(wavelengths_expected1[some])
        # #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
        # #         depths_expected.append(depths_expected1[some])
        # #     else:
        # #         pass

        # # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
        # # count_range = np.array(count_range, dtype = int)
        # # print(count_range)
        # # vgrid = np.linspace(-21,18,48)
        # # try: ccf = fits.open(file.replace('e2ds', 'ccf_K5'))
        # # except: ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
        # # rv = ccf[0].header['ESO DRS CCF RV']


        # # for line in count_range:
        # #     mid_wave = wavelengths_expected[line]
        # #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
        # #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
        # #     prof_wavelength_grid = wavelength_grid[id]
        # #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
        # #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
        # #     # id = tuple([prof_v_grid<-0.99])
        # #     # plt.figure()
        # #     # plt.plot(prof_wavelength_grid, prof)
        # #     # plt.show()
        # #     flux_grid[id] = prof

        # # # plt.figure()
        # # # plt.plot(wavelength_grid, flux_grid)
        # # # plt.show()

        # # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
        # # poly = np.poly1d(coeffs*fluxes[0])
        # # fit = poly(wavelengths)

        # # wavelengths = wavelength_grid
        # # fluxes = flux_grid * fit

        # # plt.figure()
        # plt.plot(wavelengths, fluxes)
        # plt.show()

        # find overlapping regions 
        # last_wavelengths = wave[order-1]
        # next_wavelengths = wave[order+1]
        # last_spec = spec[order-1]
        # next_spec = spec[order+1]
        # last_error = flux_error[order-1]
        # next_error = flux_error[order+1]
        # min_overlap = np.min(wavelengths)
        # max_overlap = np.max(wavelengths)

        
        # idx_ = tuple([wavelengths>min_overlap])
        # last_idx = np.logical_and(last_wavelengths>min_overlap, last_wavelengths<max_overlap)
        # next_idx = np.logical_and(next_wavelengths>min_overlap, next_wavelengths<max_overlap)
        
        # overlap = np.array(([list(last_wavelengths[last_idx]), list(last_spec[last_idx]), list(last_error[last_idx])], [list(next_wavelengths[next_idx]), list(next_spec[next_idx]), list(next_error[next_idx])]))
        
        # # overlap[0, 0] = list(last_wavelengths[last_idx])
        # # overlap[0, 1] = list(last_spec[last_idx])
        # # overlap[1, 0] = list(next_wavelengths[next_idx])
        # # overlap[1, 1] = list(next_spec[next_idx])

        # print(overlap)
        # plt.figure()
        # plt.plot(wavelengths, fluxes)
        # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
        # plt.show()
        # wavelengths = wavelengths[idx]
        # fluxes = fluxes[idx]
        # flux_error_order = flux_error_order[idx]

        if masking == 'masked':
            ## I've just been updating as I go through so it's not complete
            #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
            masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
            min_waves_mask = np.array(masks_csv[:,0])
            max_waves_mask = np.array(masks_csv[:,1])
            masks = []
            for mask_no in range(len(min_waves_mask)):
                masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

            masked_waves=[]

            for mask in masks:
                #print(np.max(mask), np.min(mask))
                idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                #print(flux_error_order[idx])
                flux_error_order[idx] = 10000000000000000000
                #print(idx)
                if len(wavelengths[idx])>0:
                    masked_waves.append(wavelengths[idx])

            #masks = []
            ### allows extra masking to be added ##

            plt.figure('masking')
            plt.plot(wavelengths, fluxes)
            if len(masked_waves)>0:
                for masked_wave in masked_waves:
                    plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
            #plt.show()

            #response = input('Are there any regions to be masked? y or n: ')
            response = 'y'
            if response == 'y':
                '''
                print('Take note of regions.')
                plt.figure('masking')
                plt.plot(wavelengths, fluxes)
                plt.show()
                '''
                #response1 = int(input('How many regions to mask?: '))
                response1 = 0
                for i in range(response1):
                    min_wave = float(input('Minimum wavelength of region %s: '%i))
                    max_wave = float(input('Maximum wavelength of region %s: '%i))
                    masks.append([min_wave, max_wave])
                masked_waves=[]
                #print(masks)
                for mask in masks:
                    #print(np.max(mask), np.min(mask))
                    idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                    #print(flux_error_order[idx])
                    flux_error_order[idx] = 10000000000000000000

                    if len(wavelengths[idx])>0:
                        masked_waves.append(wavelengths[idx])

                plt.figure('Telluric masking')
                plt.title('Spectrum - telluric masking')
                plt.plot(wavelengths, fluxes)
                # print(masked_waves)
                for masked_wave in masked_waves:
                    plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                #print('new version')
                plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))
                plt.ylabel('flux')
                plt.xlabel('wavelength')

                '''
                plt.figure('Spectrum')
                plt.title('Original Spectrum with errors')
                plt.errorbar(wavelengths, fluxes, yerr=flux_error_order, ecolor = 'k' )
                plt.ylabel('flux')
                plt.xlabel('wavelength')

                plt.show()
                '''
            if response == 'n':
                print('yay!')


        elif masking == 'unmasked':
            masked_waves = []
            masked_waves = np.array(masked_waves)

        else: print('WARNING - masking not catching - must be either "masked" or "unmasked"')

        hdu.close()

    #flux_error_order = (flux_error_order)/(np.max(fluxes)-np.min(fluxes))
    #print('flux error: %s'%flux_error_order)
    #fluxes = (fluxes - np.min(fluxes))/(np.max(fluxes)-np.min(fluxes))
    #idx = tuple([fluxes>0])
    '''
    plt.figure('Normalised Spectrum')
    plt.title('Normalised Spectrum with errors')
    plt.errorbar(wavelengths, fluxes, yerr=flux_error_order, ecolor = 'k' )
    plt.ylabel('flux')
    plt.xlabel('wavelength')

    plt.figure('Normalised Spectrum returned')
    plt.title('Normalised Spectrum returned with errors')
    plt.errorbar(wavelengths[idx], fluxes[idx], yerr=flux_error_order[idx], ecolor = 'k' )
    plt.ylabel('flux')
    plt.xlabel('wavelength')

    plt.show()
    '''
    ## telluric correction
    # tapas = fits.open('/Users/lucydolan/Starbase/tapas_000001.fits')
    # tapas_wvl = (tapas[1].data["wavelength"]) * 10.0
    # tapas_trans = tapas[1].data["transmittance"]
    # tapas.close()
    # brv=header['ESO DRS BERV']
    # tapas_wvl = tapas_wvl[::-1]/(1.+brv/2.99792458e5)
    # tapas_trans = tapas_trans[::-1]

    # background = upper_envelope(tapas_wvl, tapas_trans)
    # f = interp1d(tapas_wvl, tapas_trans / background, bounds_error=False)

    # plt.figure('telluric spec and real spec')
    # plt.plot(wavelengths, continuumfit(wavelengths, fluxes, 3))
    # plt.plot(wavelengths, f(wavelengths))
    # plt.show()
    
    # plt.figure()
    # plt.plot(tapas_wvl, tapas_trans)
    # plt.show()
    # print('overlap accounted for')
    overlap = []

    return np.array(fluxes), np.array(wavelengths), np.array(flux_error_order), sn, np.median(wavelengths), np.zeros(wavelengths.shape), overlap ## for just LSD



from astropy.io import fits
import ACID_code_v2.ACID as acid
import numpy as np
import matplotlib.pyplot as plt
import glob

def quickstart():
    spec_file = fits.open('../example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum

    linelist = '../example/example_linelist.txt' # Insert path to line list

    # choose a velocity grid for the final profile(s)
    deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
    velocities = np.arange(-25, 25, deltav)

    # run ACID function
    result = acid.ACID([wavelength], [spectrum], [error], linelist, [sn], velocities)

    # extract profile and errors
    profile = result[0, 0, 0]
    profile_error = result[0, 0, 1]

    # plot results
    plt.figure()
    plt.errorbar(velocities, profile, profile_error)
    plt.xlabel('Velocities (km/s)')
    plt.ylabel('Flux')
    plt.close('all')

def multiple_frames():

    # finds sample files in 'example directory'. Each file is a different frame.
    files = glob.glob('../example/sample_spec_*.fits')

    # create lists for wavelengths, spectra, errors and sn for all frames
    wavelengths = []
    spectra = []
    errors = []
    sns = []

    for file in files:
        spec_file = fits.open('%s'%file)

        wavelengths.append(spec_file[0].data)    # Wavelengths in Angstroms
        spectra.append(spec_file[1].data)        # Spectral Flux
        errors.append(spec_file[2].data)         # Spectral Flux Errors
        sns.append(float(spec_file[3].data))     # SN of Spectrum

    linelist = '../example/example_linelist.txt' # Insert path to line list

    # choose a velocity grid for the final profile(s)
    deltav = acid.calc_deltav(wavelengths[0])
    velocities = np.arange(-25, 25, deltav)

    # run ACID function
    result = acid.ACID(wavelengths, spectra, errors, linelist, sns, velocities)

    # plot results
    plt.figure()

    for frame in range(len(files)):
        profile = result[frame, 0, 0]
        profile_error = result[frame, 0, 1]
        plt.errorbar(velocities, profile, profile_error, label = '%s'%frame)

    plt.xlabel('Velocities (km/s)')
    plt.ylabel('Flux')
    plt.legend()
    plt.close('all')

def mulitple_orders():
    spec_file = fits.open('../example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum

    linelist = '../example/example_linelist.txt' # Insert path to line list

    # choose a velocity grid for the final profile(s)
    deltav = acid.calc_deltav(wavelength)  
    velocities = np.arange(-25, 25, deltav)

    # choose size of wavelength ranges (or chunks)
    wave_chunk = 25
    chunks_no = int(np.floor((max(wavelength)-min(wavelength))/wave_chunk))
    min_wave = min(wavelength)
    max_wave = min_wave+wave_chunk

    # create result array of shape (no. of frames, no. of chunks, 2, no. of velocity pixels)
    result = np.zeros((1, chunks_no, 2, len(velocities)))

    for i in range(chunks_no):

        # use indexing to select correct chunk of spectrum
        idx = np.logical_and(wavelength>=min_wave, wavelength<=max_wave)

        # run ACID function on specific chunk
        result = acid.ACID([wavelength[idx]], [spectrum[idx]], [error[idx]], linelist, [sn], velocities, all_frames=result, order=i)

        min_wave += wave_chunk
        max_wave += wave_chunk

    # reset min and max wavelengths
    min_wave = min(wavelength)
    max_wave = min_wave+wave_chunk

    # plot results
    plt.figure()
    for i in range(chunks_no):

        # extract profile and errors
        profile = result[0, i, 0]
        profile_error = result[0, i, 1]

        plt.errorbar(velocities, profile, profile_error, label='(%s - %sÅ)'%(min_wave, max_wave))

        min_wave += wave_chunk
        max_wave += wave_chunk

    plt.xlabel('Velocities (km/s)')
    plt.ylabel('Flux')
    plt.legend()
    plt.close('all')

quickstart()
multiple_frames()
mulitple_orders()